#ifndef ACCESS_PATTERN_H
#define ACCESS_PATTERN_H

#include "llvm/Instructions.h"
#include "llvm/Type.h"
#include "llvm/Pass.h"
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/Support/InstIterator.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Analysis/LoopInfo.h"
#include "single-access.cpp"
#include "matrix-solver.cpp"


using namespace llvm;
using namespace std;

class AccessPattern{
private:
	Value* array;
	Value* reorderedArray;
	vector<Instruction*> locationVector;

	void cleanupPartialAccess();

public:
	map<Instruction*, SingleAccess*> locationToSingleAccessMap;

	AccessPattern(Value* arr){ array = arr;};
	void addSingleAccess(Value* array, Instruction* location, 
			vector<Value*> indexVector, BasicBlock* block, LoopInfo* li);
	void printAccessRawInfo();
	
	void analysisAccessPattern();
	void printAccessMatrix();

	void makeFakeTransferMatrix();
	
	vector<int> getReorderedSizeVector();
	vector<Value*> getReorderedIndexVector(SingleAccess* singleAccess, Instruction* inst, BasicBlock* block);

	vector<vector<int> > transferMatrix;
	
	bool isPartialAccess(SingleAccess* singleAccess);
	SingleAccess* findSingleAccessFor(Instruction* inst);
	
	void addReorderedArray(Value* array){reorderedArray = array;};
	Value* getReorderedArray(){return reorderedArray;};
	
	void calculateTransferMatrix();
};

void AccessPattern::calculateTransferMatrix(){
	vector<vector<int> > allMatrix;
	vector<int> allIterationCount;
	vector<float> allIterationCountNormalized;
	float countSum = 0.0;
	for (map<Instruction*, SingleAccess*>::iterator i=locationToSingleAccessMap.begin(); 
			i!=locationToSingleAccessMap.end(); i++){
		SingleAccess* singleAccess = i->second;	
		
		allMatrix.push_back(singleAccess->getMatrix());
		allIterationCount.push_back(singleAccess->getIterationCount());
		countSum += (float)singleAccess->getIterationCount();
		cout<<"weight: "<<singleAccess->getIterationCount()<<endl;
	}
	for (int i = 0; i < allIterationCount.size(); i++)
		allIterationCountNormalized.push_back((float)allIterationCount[i]/countSum);
	transferMatrix = (new MatrixSolver())->solve(allMatrix, -5, 5, allIterationCountNormalized);
	cout<<"displaying results..."<<endl;
	(new MatrixSolver())->printMatrix(transferMatrix);
	//reordering columns
	for(int i=0; i<transferMatrix.size(); i++){
		int col_size = transferMatrix[i].size()-1;
		for(int j=0; j<=col_size/2; j++){
			int temp = transferMatrix[i][j];
			transferMatrix[i][j] = transferMatrix[i][col_size-j];
			transferMatrix[i][col_size-j] = temp;
		}
	}
	//reordering rows
	int row_size = transferMatrix.size()-1;
	for(int i=0; i<=row_size/2; i++){
		vector<int> temp  = transferMatrix[i];
		transferMatrix[i] = transferMatrix[row_size-i];
		transferMatrix[row_size-i] = temp;
	}
	
	if (transferMatrix.size() == 0){
		cout<<"use default result..."<<endl;
		makeFakeTransferMatrix();
	}
	cout<<"displaying results..."<<endl;
	(new MatrixSolver())->printMatrix(transferMatrix);
}

void AccessPattern::makeFakeTransferMatrix(){
	int dimension = (new Util())->getDimForArray(array);
	for (int i = 0; i < dimension; i++){
		vector<int> row;
		for (int j = 0; j < dimension; j++){
			if (i == j)
				row.push_back(1);
			else
				row.push_back(0);
		}
		transferMatrix.push_back(row);
	}
}

vector<int> AccessPattern::getReorderedSizeVector(){
	vector<int> sizeVector;
	vector<int> originalSizeVector = (new Util())->getSizeVectorForArray(array);
	for (int i = 0; i < (new Util())->getDimForArray(array); i++){
		if ((new Util())->getDimForArray(array) >= 3)
			sizeVector.push_back(100);
		else
			sizeVector.push_back(1000);
	}
	return sizeVector;
}

vector<Value*> AccessPattern::getReorderedIndexVector(SingleAccess* singleAccess, Instruction* inst, BasicBlock* block){
	return singleAccess->getReorderedIndexVector(transferMatrix, inst, block);
}

SingleAccess* AccessPattern::findSingleAccessFor(Instruction* inst){
	return locationToSingleAccessMap[inst];
}

bool AccessPattern::isPartialAccess(SingleAccess* singleAccess){
	int dimension = (new Util())->getDimForArray(array);
	if (singleAccess->getIndexVector().size() < dimension)
		return true;
	else
		return false;
}

void AccessPattern::cleanupPartialAccess(){
	int dimension = (new Util())->getDimForArray(array);
	vector<Instruction*> keysToRemove;
	errs()<<"DIM: "<<dimension<<"\n";
	for (map<Instruction*, SingleAccess*>::iterator i = locationToSingleAccessMap.begin();
			i != locationToSingleAccessMap.end(); i++){
		Instruction* inst = i->first;
		SingleAccess* singleAccess = i->second;
		if (isPartialAccess(singleAccess)){
			keysToRemove.push_back(inst);
		}
	}
	for (int i = 0; i < keysToRemove.size(); i++)
		locationToSingleAccessMap.erase(keysToRemove[i]);
}

void AccessPattern::analysisAccessPattern(){
	cleanupPartialAccess();
	for (map<Instruction*, SingleAccess*>::iterator i = locationToSingleAccessMap.begin();
			i != locationToSingleAccessMap.end(); i++){
		Instruction* inst = i->first;
		SingleAccess* singleAccess = i->second;
		
		singleAccess->updateIndVarMatrix();
	}
}

void AccessPattern::addSingleAccess(
		Value* arr, Instruction* loc,
		vector<Value*> indexVec, BasicBlock* bb, LoopInfo* li){
	if (arr != array){
		errs()<<"unmatched access\n";
		return;
	}
        errs()<<"BEGIN=========insertSingleAccess=============\n";
        errs()<<"location: "<<*loc<<"\n";
        errs()<<"array: "<<*arr<<"\n";
        for (int i = 0; i < indexVec.size(); i++){
                Value* v = indexVec[i];
                errs()<<"index: "<<*v<<"\n";
        }
        errs()<<"block: "<<bb->getName()<<"\n";
        errs()<<"END=========insertSingleAccess=============\n";

        SingleAccess* singleAccess = new SingleAccess(arr, loc, indexVec, bb, li);
        locationToSingleAccessMap[loc] = singleAccess;
}

void AccessPattern::printAccessRawInfo(){
	for (map<Instruction*, SingleAccess*>::iterator i = locationToSingleAccessMap.begin();
			i != locationToSingleAccessMap.end(); i++){
		Instruction* inst = i->first;
		errs()<<"LOC: "<<*inst<<"\n";
		SingleAccess* singleAccess = i->second;
		for (int i = 0; i < singleAccess->getIndexVector().size(); i++){
			Value* index = singleAccess->getIndexVector()[i];
			errs()<<"INDEX: "<<*index<<"\n";
		}
	}
}

void AccessPattern::printAccessMatrix(){
	for (map<Instruction*, SingleAccess*>::iterator i = locationToSingleAccessMap.begin();
			i != locationToSingleAccessMap.end(); i++){
		Instruction* inst = i->first;
		SingleAccess* singleAccess = i->second;
		
		singleAccess->printIndVarMatrix();
	}
}


#endif
