#include "llvm/Pass.h"
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/Support/InstIterator.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Instructions.h"
#include "llvm/Type.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/ADT/SmallVector.h"
#include "llvm/Argument.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/IRBuilder.h"
#include "llvm/InstrTypes.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Type.h"
#include "llvm/DerivedTypes.h"
#include "llvm/Constants.h"
#include "llvm/Bitcode/ReaderWriter.h"

#include <ostream>
#include <fstream>
#include <iostream>

#include "access-pattern.cpp"

using namespace llvm;
using namespace std;

namespace {

class ArrayReorderingV3 : public ModulePass {

public:

static char ID;

map<Value*, AccessPattern*> arrayToAccessPatternMap;

ArrayReorderingV3() : ModulePass(ID) { }

void getAnalysisUsage(AnalysisUsage &AU) const {
 AU.setPreservesAll();
 AU.addRequired<LoopInfo>();
} 

bool inspectFunction(Function &F) {
	for (Function::iterator FI = F.begin(), FE = F.end(); FI != FE; ++FI) {
		BasicBlock* block = FI;
		errs()<<block->getName()<<"\n";
		for (BasicBlock::iterator BI = block->begin(), BE = block->end(); BI != BE; ++BI) {
			Instruction* inst = BI;
			errs()<<*inst<<"\n";
			if (AllocaInst::classof(inst)){
				AllocaInst* allocaInst = (AllocaInst*)inst;
				Type* type = allocaInst->getAllocatedType();
				if (type->isArrayTy()){
					errs()<<"====== initialized an array type\n";
				}
			}
			if (GetElementPtrInst::classof(inst)){
				errs()<<"====== get array index"<<"\n";
			}
		}
	}
	return false;
}

bool analyzeFunction(Function &F, LoopInfo* loopInfo) {
	for (Function::iterator FI = F.begin(), FE = F.end(); FI != FE; ++FI) {
		BasicBlock* block = FI;
		errs()<<block->getName()<<"\n";
		for (BasicBlock::iterator BI = block->begin(), BE = block->end(); BI != BE; ++BI) {
			Instruction* inst = BI;
			errs()<<*inst<<"\n";
			if (AllocaInst::classof(inst)){
				AllocaInst* allocaInst = (AllocaInst*)inst;
				Type* type = allocaInst->getAllocatedType();
				if (type->isArrayTy()){
					errs()<<"====== initialized an array type\n";
					Value* array = (Value*)allocaInst;
					AccessPattern* accessPattern = new AccessPattern(array);
					arrayToAccessPatternMap[array] = accessPattern;
				}
			}
			if (GetElementPtrInst::classof(inst)){
				errs()<<"====== get array index"<<"\n";
				vector<Value*> indexVector = (new Util())->getIndexVectorForGetElementPtr(inst);
				Value* array = (new Util())->getArrayForGetElementPtr(inst);
				AccessPattern* accessPattern = arrayToAccessPatternMap[array];
				accessPattern->addSingleAccess(array, inst, indexVector, block, loopInfo);
				arrayToAccessPatternMap[array] = accessPattern;
			}
		}
	}
	return false;
}

bool analyzeIterationInfo(Function &F, LoopInfo* loopInfo) {
	for (map<Value*, AccessPattern*>::iterator i=arrayToAccessPatternMap.begin(); 
			i!=arrayToAccessPatternMap.end(); i++){		
		errs()<<"\n";
		Value* v = i->first;
		AccessPattern* accessPattern = i->second;
		for (map<Instruction*, SingleAccess*>::iterator i = accessPattern->locationToSingleAccessMap.begin();
			i != accessPattern->locationToSingleAccessMap.end(); i++){
			Instruction* inst = i->first;
			SingleAccess* singleAccess = i->second;
			vector<Value*> indVarChain = singleAccess->getIndVarChain();
			for (int i = 0; i < indVarChain.size(); i++){
				Value* indVar = indVarChain[i];
				int loopIterationCount = getLoopIterationCountForIndVar(F, loopInfo, indVar);
				//errs()<<"&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&"<<*indVar<<"\t"<<loopIterationCount<<"\n";
				singleAccess->indVarToIterationCount[indVar] = loopIterationCount;
			}
		}
	}
	return false;
}

int getLoopIterationCountForIndVar(Function &F, LoopInfo* loopInfo, Value* indVar){
	for (Function::iterator FI = F.begin(), FE = F.end(); FI != FE; ++FI) {
		BasicBlock* block = FI;
		//errs()<<block->getName()<<"\n";
		for (BasicBlock::iterator BI = block->begin(), BE = block->end(); BI != BE; ++BI) {
			Instruction* inst = BI;
			//errs()<<*inst<<"\n";
			if (ICmpInst::classof(inst)){
				errs()<<*inst<<"\n";
				ICmpInst* iCmpInst = (ICmpInst*)inst;
				Value* v0 = iCmpInst->getOperand(0);
				Value* v1 = iCmpInst->getOperand(1);
				LoadInst* loadInst = (LoadInst*)v0;
				Value* indv = loadInst->getOperand(0);
				if (indv == indVar){
					//errs()<<*v0<<"\t"<<*v1<<"\t"<<*indv<<"\n";
					int val = (int)((ConstantInt*)v1)->getSExtValue();
					return val;
				}
			}
		}
	}
	return -1;
}


bool insertingReorderedArray(Function &F, LoopInfo* loopInfo){
	vector<Instruction*> instToRemove;
	for (Function::iterator FI = F.begin(), FE = F.end(); FI != FE; ++FI) {
		BasicBlock* block = FI;
		errs()<<block->getName()<<"\n";
		for (BasicBlock::iterator BI = block->begin(), BE = block->end(); BI != BE; ++BI) {
			Instruction* inst = BI;
			errs()<<*inst<<"\n";
			if (AllocaInst::classof(inst)){
				AllocaInst* allocaInst = (AllocaInst*)inst;
				Type* type = allocaInst->getAllocatedType();
				if (type->isArrayTy()){
					errs()<<"====== initialized an array type\n";
					Value* array = (Value*) allocaInst;
					if (arrayToAccessPatternMap.count(array) == 1){
						AccessPattern* accessPattern = arrayToAccessPatternMap[array];
						vector<int> newSizeVector = accessPattern->getReorderedSizeVector();
						AllocaInst* newInst = 
								(new Util())->getArrayInitialInstruction(newSizeVector, allocaInst);
						(new Util())->insertInst(block, inst, newInst);
						//block->getInstList().insert(inst, newInst);
						accessPattern->addReorderedArray((Value*)newInst);
					}
				}
			}
			if (GetElementPtrInst::classof(inst)){
				errs()<<"====== get array index"<<"\n";
				vector<Value*> indexVector = (new Util())->getIndexVectorForGetElementPtr(inst);
				Value* array = (new Util())->getArrayForGetElementPtr(inst);
				if (arrayToAccessPatternMap.count(array) == 1){
					AccessPattern* accessPattern = arrayToAccessPatternMap[array];
					SingleAccess* singleAccess = accessPattern->findSingleAccessFor(inst);
					if (singleAccess != NULL && !accessPattern->isPartialAccess(singleAccess)){
						vector<Value*> newIndexVector = 
								accessPattern->getReorderedIndexVector(singleAccess, inst, block);
						vector<Instruction*> instructionVector = 
								(new Util())->getArrayAccessInstructionVector(
										newIndexVector, accessPattern->getReorderedArray());
						for (int i = 0; i < instructionVector.size(); i++){
							(new Util())->insertInst(block, inst, instructionVector[i]);
						}
						Value* finalNewValue = (Value*)(instructionVector[instructionVector.size()-1]);
						((Value*)inst)->replaceAllUsesWith(finalNewValue);
						errs()<<"REPLACING "<<*inst<<" BY "<<*finalNewValue<<"\n";
						instToRemove.push_back(inst);
						continue;
					}
				}
				Value* parentArray = (new Util())->getArrayForGetElementPtr(inst);
				if (arrayToAccessPatternMap.count(parentArray) != 0){
					errs()<<"ABOUT TO ERASE: "<<*inst<<"\n";
					instToRemove.push_back(inst);
				}
			}
		}
	}
	/*for (int i = 0; i < instToRemove.size(); i++){
		errs()<<"ERASING: "<<*(instToRemove[i])<<"\n";
		instToRemove[i]->eraseFromParent();
	}*/
	return false;
}

bool cleaningupOldArray(Function &F, LoopInfo* loopInfo){
	return false;
}

bool runOnModule(Module& M) {
	
	for (Module::iterator MI = M.begin(), ME = M.end(); MI != ME; ++MI) {
		Function &F = *MI;
		LoopInfo* loopInfo = &getAnalysis<LoopInfo>(F);
		//LoopInfo* loopInfo;
		analyzeFunction(F, loopInfo);
		analyzeIterationInfo(F, loopInfo);
	}
	for (map<Value*, AccessPattern*>::iterator i=arrayToAccessPatternMap.begin(); 
			i!=arrayToAccessPatternMap.end(); i++){		
		errs()<<"\n";
		Value* v = i->first;
		AccessPattern* accessPattern = i->second;
		errs()<<"ANALYZING ARRAY: "<<*v<<"\n";
		errs()<<"PRINTING OUT ALL ACCESS RAW INFO\n";
		accessPattern->printAccessRawInfo();
		
		errs()<<"ANALYZING ACCESS MATRIX\n";
		accessPattern->analysisAccessPattern();
		errs()<<"PRINTING OUT ACCESS MATRIX\n";
		accessPattern->printAccessMatrix();

		errs()<<"GENERATING NEW ARRAY\n";
		accessPattern->calculateTransferMatrix();
		//accessPattern->makeFakeTransferMatrix();		
		
	}
	
	errs()<<"INSERTING NEW ARRAY TO ORIGINAL CODE\n";
	for (Module::iterator MI = M.begin(), ME = M.end(); MI != ME; ++MI) {
		Function &F = *MI;
		LoopInfo* loopInfo = &getAnalysis<LoopInfo>(F);
		insertingReorderedArray(F, loopInfo);
	}
	errs()<<"INSERTING FINISHES\n";
	
	errs()<<"CLEANING UP OLD ARRAYS\n";
	for (Module::iterator MI = M.begin(), ME = M.end(); MI != ME; ++MI) {
		Function &F = *MI;
		LoopInfo* loopInfo = &getAnalysis<LoopInfo>(F);
		cleaningupOldArray(F, loopInfo);
	}
	errs()<<"CLEANING UP FINISHES\n";

	errs()<<"\n";

	errs()<<"PROCESSED RESULTS\n";
	for (Module::iterator MI = M.begin(), ME = M.end(); MI != ME; ++MI) {
		Function &F = *MI;
		inspectFunction(F);
	}
	errs()<<"ALL FINISHES\n";
	return true;
}

private:
};

char ArrayReorderingV3::ID = 0;
RegisterPass<ArrayReorderingV3> X("cd-ar3",
    "15745 ArrayReorderingV3");

}
