#ifndef UTIL_H
#define UTIL_H

#include "llvm/Instructions.h"
#include "llvm/Type.h"
#include "llvm/Analysis/LoopInfo.h"

using namespace llvm;
using namespace std;


class Util{
public:
	vector<Loop*> getLoopChain(Loop* loop){
		vector<Loop*> loopChain;
		while (loop != NULL){
			loopChain.push_back(loop);
			loop = loop->getParentLoop();
		}
		return loopChain;
	};
	
	Value* getIndVarFromLoop(Loop* loop){
		BasicBlock* header = loop->getHeader();
		BasicBlock* loopIncBlock;
		for (pred_iterator PI = pred_begin(header), E = pred_end(header); PI != E; ++PI){
			loopIncBlock = *PI;
			break; //only need to consider the first pred which we assume to be loop.inc
		}
		for (BasicBlock::iterator BI = loopIncBlock->begin(), BE = loopIncBlock->end(); BI != BE; ++BI) {
			Instruction* inst = BI;
			if (LoadInst::classof(inst)){
				Value* indVar = inst->getOperand(0);
				return indVar;
			}
			else{
				return NULL;
			}
		}
	};

	vector<int> getLinearParamForIndex(Value* index, vector<Value*> indVarChain){
		vector<int> linearParams;
		for (int i = 0; i < indVarChain.size(); i++)
			linearParams.push_back(0);
		Instruction* inst = (Instruction*)index;
		if (inst == NULL)
			return linearParams;
		if (BinaryOperator::classof(inst)){
			BinaryOperator *bin = (BinaryOperator*)inst;
			Instruction::BinaryOps op = bin->getOpcode();
			Value *v1 = bin->getOperand(0);
			Value *v2 = bin->getOperand(1);
			vector<int> paramsLeft = getLinearParamForIndex(v1, indVarChain);
			vector<int> paramsRight = getLinearParamForIndex(v2, indVarChain);
			switch(op){
				case Instruction::Add: {
					for (int i = 0; i < paramsLeft.size(); i++){
						linearParams[i] = paramsLeft[i]+paramsRight[i];
					}
					break;
				}
				case Instruction::Sub: {
					for (int i = 0; i < paramsLeft.size(); i++){
						linearParams[i] = paramsLeft[i]-paramsRight[i];
					}
					break;
				}
				case Instruction::Mul: {
					if (ConstantInt::classof(v1) && !ConstantInt::classof(v2)){
						int val = (int)((ConstantInt*)v1)->getSExtValue();
						for (int i = 0; i < paramsLeft.size(); i++){
							linearParams[i] = paramsRight[i]*val;
						}
					}
					else if (ConstantInt::classof(v2) && !ConstantInt::classof(v1)){
						int val = (int)((ConstantInt*)v2)->getSExtValue();
						for (int i = 0; i < paramsLeft.size(); i++){
							linearParams[i] = paramsLeft[i]*val;
						}
					}
					break;
				}
			}
			//errs()<<"RECURSIVE: "<<*index<<" ["<<*v1<<","<<*v2<<"] "<<" ["<<paramsLeft[0]<<","<<paramsLeft[1]<<"] "<<" ["<<paramsRight[0]<<","<<paramsRight[1]<<"] "<<" ["<<linearParams[0]<<","<<linearParams[1]<<"]\n";
		}
		else if (LoadInst::classof(inst)){
			Value* v = inst->getOperand(0);
			for (int i = 0; i < indVarChain.size(); i++){
				Value* indVar = indVarChain[i];
				if (indVar == v){	
					linearParams[i] = 1;
				}
			}
		}
		return linearParams;
	};

	vector<Value*> getIndexVectorForGetElementPtr(Instruction* inst){
		vector<Value*> indexVector;
		while (GetElementPtrInst::classof(inst)){
			Value* array = inst->getOperand(0);
			Value* index = inst->getOperand(2);
			indexVector.push_back(index);
			inst = (Instruction*)array;
		}
		return indexVector;
	};

	Value* getArrayForGetElementPtr(Instruction* inst){
		vector<Value*> indexVector;
		Value* array;
		while (GetElementPtrInst::classof(inst)){
			array = inst->getOperand(0);
			inst = (Instruction*)array;
		}
		return array;
	};

	AllocaInst* getArrayInitialInstruction(vector<int> sizeVector, AllocaInst* originalInst){
		//first, get nested depth and lowest-level element type
		int dim = 0;
		Type* elementType = originalInst->getAllocatedType();
		while (true){
			dim++;
			elementType = ((ArrayType*)elementType)->getElementType();
			if (!elementType->isArrayTy())
				break;
		}
		if (dim != sizeVector.size()){
			errs()<<"unmatched size\n";
			return originalInst;
		}
		ConstantInt* index;
		ArrayType* arrayType;
		for (int i = 0; i < sizeVector.size(); i++){
			if (i == 0){
				arrayType = ArrayType::get(elementType, sizeVector[i]);
				index = ConstantInt::get(Type::getInt32Ty(originalInst->getContext()), 1, false);
			} else {				
				arrayType = ArrayType::get(arrayType, sizeVector[i]);
				index = ConstantInt::get(Type::getInt32Ty(originalInst->getContext()), 0, false);
			}
		}
		AllocaInst* newInst = new AllocaInst(arrayType, originalInst->getArraySize(), "NewArray");
		newInst->setAlignment(4);
		return newInst;
	};

	//return index retrival from outmost to innermost loop
	vector<Instruction*> getArrayAccessInstructionVector(vector<Value*> indexVector, Value* array){
		vector<Instruction*> instructionVector;
		for (int i = indexVector.size()-1; i >= 0; i--){
			vector<Value*> idx_vector;
			ConstantInt* index = ConstantInt::get(Type::getInt32Ty(((Instruction*)array)->getContext()), 0, false);
			idx_vector.push_back(index);
			//index = ConstantInt::get(Type::getInt32Ty(BI->getContext()), indexVector[i], false);
			idx_vector.push_back(indexVector[i]);
			GetElementPtrInst* newGetInst = GetElementPtrInst::CreateInBounds(array, ArrayRef<Value*>(idx_vector), "NewIndex");
			Instruction* newInst = (Instruction*)newGetInst;
			instructionVector.push_back(newInst);
			array = (Value*)newGetInst;
		}
		return instructionVector;
	};
	
	int getDimForArray(Value* array){
		int dim = 0;
		Type* elementType = ((AllocaInst*)array)->getAllocatedType();
		while (true){
			dim++;
			elementType = ((ArrayType*)elementType)->getElementType();
			if (!elementType->isArrayTy())
				break;
		}
		return dim;
	}

	vector<int> getSizeVectorForArray(Value* array){
		vector<int> sizeVector;
		//ArrayType *a = cast<ArrayType>(((ArrayType*)array)->getElementType());
		//errs()<<"&&&&&&&&&&&&&&&&&&&&&&&&"<<*array<<"\t\t"<<((ArrayType*)array)->getNumElements()<<"\n";
		return sizeVector;
	}
	
	void insertInst(BasicBlock* block, Instruction* location, Instruction* newInst){
		block->getInstList().insert(location, newInst);
		errs()<<"--------ADD INSTRUCTION"<<*newInst<<" BEFORE "<<*location<<"\n";
	}
};

#endif
